#ifndef __NETMESSAGE_H__
#define __NETMESSAGE_H__

class NetMessage;

enum NETMESSAGES
{
    NM_Invalid = 0,

    // Reliable UDP Messages.
    NM_Ack,                 // 1
    NM_KeepAlive,           // 2

    // NetCode'ish Messages.
    NM_JoinServer = 100,    // 100
    NM_Welcome,             //
    NM_LeaveServer,         //
    NM_Goodbye,             //

    // Game Messages.
    NM_AddPlayer,           // player request to be added to the game.
    NM_PlayerJoined,        // server response telling everyone a player was added.
    NM_RemovePlayer,        // ... not sure why I did this ...
    NM_PlayerLeft,          // sent by players when they die, and sent by server when a client drops/times-out

    NM_RequestGameState,    // sent by clients when they first join.
    NM_GameInfo,            // server will sent if requested, or when a round ended to set up the next round.
    NM_StartGame,           // start a round of play.
    NM_RoundEnded,          // notify the clients that a round ended... could probably not be sent, letting gameinfo do this work.

    NM_SetOnFire,           // when hit by a bomb.
    NM_PlayerMove,          // sent 4 times per second //////10 times/second/player, should probably reduce the freq.
    NM_RequestBomb,         // tell the server we're placing a bomb, don't light the wick 'til the server confirms it.
    NM_DropBomb,            // sent to clients telling them about the bomb and when to pop it.

    NM_MoveElevator,        // sent by host when a player steps on an elevator.
    NM_GrabPowerUp,         // sent by host when a player steps on a power up.

    NM_BombFlood,           // sent by host when game is dragging on.

    NM_PlayerInfo,          // sent 4 times per second to each client.
};

NetMessage* MakeMessageCopy(NetMessage* nm);

class NetMessage
{
public:
    // this stuff really shouldn't be sent on the network,
    // but it's tiny so I won't bother fixing it.
    u8 reliable;
    u8 refcount;
    u8 forward; // this could be set based on the message type at the other end.

    // this stuff should
    u8 player;
    u16 id;
    u16 padding;
    NETMESSAGES type;

    NetMessage() : reliable(0), refcount(1), forward(0), player(255), id(0), type(NM_Invalid) {}
    virtual ~NetMessage() {}

    virtual void Send(int clienttosendto = -1) = 0;
    virtual void Send(int size, int clienttosendto);
    void CopyFromBase(NetMessage* nm)
    {
        reliable = nm->reliable;

        player = nm->player;
        id = nm->id;
        type = nm->type;
    }
};

class MessageAck : public NetMessage
{
public:
    MessageAck()
    {
        type = NM_Ack;
    }
    virtual ~MessageAck() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
};

class MessageKeepAlive : public NetMessage
{
public:
    MessageKeepAlive()
    {
        type = NM_KeepAlive;
    }
    virtual ~MessageKeepAlive() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
};

class MessageJoinServer : public NetMessage
{
public:
    MessageJoinServer()
    {
        type = NM_JoinServer;
        reliable = 1;
    }
    virtual ~MessageJoinServer() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageJoinServer* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageWelcome : public NetMessage
{
public:
    int machinenumber;

    MessageWelcome()
    {
        type = NM_Welcome;
        reliable = 1;
    }
    virtual ~MessageWelcome() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageWelcome* nm)
    {
        machinenumber = nm->machinenumber;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageLeaveServer : public NetMessage
{
public:
    MessageLeaveServer()
    {
        type = NM_LeaveServer;
        reliable = 1;
    }
    virtual ~MessageLeaveServer() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageLeaveServer* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageGoodbye : public NetMessage
{
public:
    MessageGoodbye()
    {
        type = NM_Goodbye;
        reliable = 1;
    }
    virtual ~MessageGoodbye() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageGoodbye* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageAddPlayer : public NetMessage
{
public:
    char name[12];

    MessageAddPlayer()
    {
        type = NM_AddPlayer;
        reliable = 1;
    }
    virtual ~MessageAddPlayer() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageAddPlayer* nm)
    {
        strcpy(name, nm->name);
        CopyFromBase((NetMessage*)nm);
    }
};

class MessagePlayerJoined : public NetMessage
{
public:
    int score;
    char name[12];
    int machinewhocontrolsit;

    MessagePlayerJoined()
    {
        type = NM_PlayerJoined;
        reliable = 1;
    }
    virtual ~MessagePlayerJoined() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessagePlayerJoined* nm)
    {
        score = nm->score;
        strcpy(name, nm->name);
        machinewhocontrolsit = nm->machinewhocontrolsit;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageRemovePlayer : public NetMessage
{
public:
    MessageRemovePlayer()
    {
        type = NM_RemovePlayer;
        reliable = 1;
    }
    virtual ~MessageRemovePlayer() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageRemovePlayer* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessagePlayerLeft : public NetMessage
{
public:
    MessagePlayerLeft()
    {
        type = NM_PlayerLeft;
        reliable = 1;
        forward = 1;
    }
    virtual ~MessagePlayerLeft() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessagePlayerLeft* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageRequestGameState : public NetMessage
{
public:
    MessageRequestGameState()
    {
        type = NM_RequestGameState;
        reliable = 1;
    }
    virtual ~MessageRequestGameState() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageRequestGameState* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageGameInfo : public NetMessage
{
public:
    // if you're going to somewhat optimize a packet,
    // make sure it's the one that's sent only once per round... ack. :)

    // left 4 bits are powerups, right 4 bits are blocks
    u8 blocks[NUM_MAPS][MAP_WIDTH][MAP_HEIGHT];
    //u8 powerups[NUM_MAPS][MAP_WIDTH][MAP_HEIGHT];
    int framecount;
    int roundstarted;

    MessageGameInfo()
    {
        type = NM_GameInfo;
        reliable = 1;
    }
    virtual ~MessageGameInfo() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageGameInfo* nm)
    {
        memcpy(blocks, nm->blocks, sizeof(blocks));
        //memcpy(powerups, nm->powerups, sizeof(powerups));
        framecount = nm->framecount;
        roundstarted = nm->roundstarted;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageStartGame : public NetMessage
{
public:
    MessageStartGame()
    {
        type = NM_StartGame;
        reliable = 1;
    }
    virtual ~MessageStartGame() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageStartGame* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageRoundEnded : public NetMessage
{
public:
    MessageRoundEnded()
    {
        type = NM_RoundEnded;
        reliable = 1;
    }
    virtual ~MessageRoundEnded() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageRoundEnded* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageSetOnFire : public NetMessage
{
public:
    MessageSetOnFire()
    {
        type = NM_SetOnFire;
        reliable = 1;
        forward = 1;
    }
    virtual ~MessageSetOnFire() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageSetOnFire* nm)
    {
        CopyFromBase((NetMessage*)nm);
    }
};

class MessagePlayerMove : public NetMessage
{
public:
    float x;
    float y;
    float z;
    int keys;

    MessagePlayerMove() : x(-1), y(-1), z(-1), keys(0)
    {
        type = NM_PlayerMove;
        forward = 0;
    }
    virtual ~MessagePlayerMove() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessagePlayerMove* nm)
    {
        x = nm->x;
        y = nm->y;
        z = nm->z;
        keys = nm->keys;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageRequestBomb : public NetMessage
{
public:
    int map;
    int tx;
    int ty;
    int range;

    MessageRequestBomb() : map(0), tx(0), ty(0), range(0)
    {
        type = NM_RequestBomb;
        reliable = 1;
        forward = 0;
    }
    virtual ~MessageRequestBomb() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageRequestBomb* nm)
    {
        map = nm->map;
        tx = nm->tx;
        ty = nm->ty;
        range = nm->range;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageDropBomb : public NetMessage
{
public:
    int map;
    int tx;
    int ty;
    int range;
    int TickToExplode;

    MessageDropBomb() : map(0), tx(0), ty(0), TickToExplode(0)
    {
        type = NM_DropBomb;
        reliable = 1;
    }
    virtual ~MessageDropBomb() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageDropBomb* nm)
    {
        map = nm->map;
        tx = nm->tx;
        ty = nm->ty;
        range = nm->range;
        TickToExplode = nm->TickToExplode;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageMoveElevator : public NetMessage
{
public:
    u8 ElevatorGroup;
    int TickToMove;

    MessageMoveElevator() : ElevatorGroup(2), TickToMove(0)
    {
        type = NM_MoveElevator;
        reliable = 1;
    }
    virtual ~MessageMoveElevator() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageMoveElevator* nm)
    {
        ElevatorGroup = nm->ElevatorGroup;
        TickToMove = nm->TickToMove;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageGrabPowerUp : public NetMessage
{
public:
    int map;
    int tx;
    int ty;

    MessageGrabPowerUp()
    {
        type = NM_GrabPowerUp;
        reliable = 1;
    }
    virtual ~MessageGrabPowerUp() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageGrabPowerUp* nm)
    {
        map = nm->map;
        tx = nm->tx;
        ty = nm->ty;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessageBombFlood : public NetMessage
{
public:
    u8 map;
    u8 side;

    MessageBombFlood() : side(0)
    {
        type = NM_BombFlood;
        reliable = 1;
    }
    virtual ~MessageBombFlood() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void CopyFrom(MessageBombFlood* nm)
    {
        map = nm->map;
        side = nm->side;
        CopyFromBase((NetMessage*)nm);
    }
};

class MessagePlayerInfo : public NetMessage
{
    // I've done some optimizing of this packet since it's sent a lot by the server to each client.
    // yet still thinking of "always worst case" and sending info of all 8 players all the time.
public:
    u8 x[MAX_PLAYERS];
    u8 y[MAX_PLAYERS];
    u8 keys[MAX_PLAYERS]; // player z is stored in bits 0-4 of keys.

    MessagePlayerInfo()
    {
        type = NM_PlayerInfo;
        reliable = 0;
        forward = 0;
    }
    virtual ~MessagePlayerInfo() {}

    virtual void Send(int clienttosendto = -1) { NetMessage::Send( sizeof(*this), clienttosendto ); }
    void Fill(int player, float nx, float ny, float nz, int nkeys)
    {
        x[player] = (u8)nx;
        y[player] = (u8)ny;
        u8 z4 = (u8)nz >> 4;
        keys[player] = ((u8)nkeys & 0xF0) | z4;
    }
    float GetX(int player)
    {
        return (float)x[player];
    }
    float GetY(int player)
    {
        return (float)y[player];
    }
    float GetZ(int player)
    {
        return (float)((keys[player] & 0x0F) << 4);
    }
    int GetKeys(int player)
    {
        return keys[player] & 0xF0;
    }

    void CopyFrom(MessagePlayerInfo* nm)
    {
        int i;
        for(i=0; i<MAX_PLAYERS; i++)
        {
            x[i] = nm->x[i];
            y[i] = nm->y[i];
            keys[i] = nm->keys[i];
        }
        CopyFromBase((NetMessage*)nm);
    }
};

#endif //__NETMESSAGE_H__
